<?php
 /**
  * @property   mysql_class $db An Instance of mysql_class (needs a replacment)
  * @author     Reindl Harald <h.reindl@thelounge.net>
  * @license    public domain
 */
 class panel_dns
 {
  public $db;

  /** @var array   $types                Liste aller unterstuetzten Eintragstypen */
  public $types        = array('A', 'CNAME', 'MX', 'SPF', 'TXT', 'NS');

  /** @var array   $blacklist            Liste nicht erlaubter Domains */
  public $blacklist    = array();

  /** @var string  $soa                  SOA- und NS-Eintrag */
  public $soa          = '';

  /** @var string  $ns2                  Zweiter Nameserver */
  public $ns2          = '';

  /** @var string  $ns3                  Dritter Nameserver */
  public $ns3          = '';

  /** @var string  $hostmaster           Technischer Verantwortlicher */
  public $hostmaster   = '';

  /** @var integer $ttl                  Time-To-Live */
  public $ttl          = 3600;

  /** @var integer $negative_ttl         Negative-Time-To-Live */
  public $negative_ttl = 3600;

  /** @var integer $refresh              Aktualisierungs-Interval */
  public $refresh      = 3600;

  /** @var integer $retry                Interval fuer neue Versuche bei Fehlern */
  public $retry        = 3600;

  /** @var integer $expire               Ablaufzeit am Slave in Sekunden wenn keine Master-Antwort */
  public $expire       = 3628800;

  /** @var string   $default_host_ip     Standard-Host einer Domain ohne Angaben */
  public $default_host_ip;

  /** @var string   $default_host_cname  Standard-CNAME einer Domain ohne Angaben */
  public $default_host_cname;

  /** @var array   $default_spf_new      Standard-SPF-Eintraege einer neuen Domain */
  public $default_spf_new;

  /** @var array    $default_spf         Liste der Standard-SPF-Hosts um den Value 'default' beim Aktualisieren von Zonen zu uebersetzen */
  public $default_spf;

  /** @var boolean $default_spf_strict   SPF-Policy bei neuen Domains */
  public $default_spf_strict = 0;

  /** @var boolean  $batch_mode          Zones werden nicht automatisch generiert sondern manuell am Ende angestossen */
  public $batch_mode = false;

  /** @var string   $sql_prefix          Praefix um das Interface fuer mehrere DNS-Server verwenden zu koennen */
  public $sql_prefix = '';

  /** @var string   $dhcp_zone           Domain bei der auch MAC-Adressen zugewiesen werden koennen */
  public $dhcp_zone = '';

  /** @var array    $client_ids          Interne Zuordnungsinformationen zu DNS-Zonen definieren */
  public $client_ids = array();

  /** @var integer  $tab_fuzz            Fuzzy-Logic-Wert fur TABS im Zone-File */
  private $tab_fuzz = 7;

  /**
   * Array um Zonen-Daten mit uebersetzten LAN-Adressen im Feld 'zone_content_lan' abzulegen
   * Defacto ein Replacement des aus der DB generierten Feldes 'zone_content'
   * Umsetzung der Config am internen DNS gesondert zu implementieren
  */
  public $lan_translation = array();

  /** Initailisierung des DNS-Systems */
  public function init()
  {
   /** Sicherstellen dass Import-Tabelle existiert */
   if(!$this->db->table_exists($this->sql_prefix . 'import'))
   {
    $this->db->create_table
    (
     $this->sql_prefix . 'import',
     array
     (
      array('name'=>'prikey'),
      array('name'=>'domain',  'type'=>'varchar', 'length'=>255),
      array('name'=>'type',    'type'=>'varchar', 'length'=>50),
      array('name'=>'name',    'type'=>'varchar', 'length'=>255),
      array('name'=>'value',   'type'=>'varchar', 'length'=>255),
      array('name'=>'comment', 'type'=>'varchar', 'length'=>255)
     ),
     /**$key_array*/array(),
     /**$drop_if_exists*/0,
     /**$execute*/1,
     /**$target_srvtyp*/-1,
     /**$temp*/0
    );
   }
   /** Sicherstellen dass die Meta-Tabelle einen Datensatz hat */
   $row = $this->db->fetch_row($this->db->query('select count(*) from ' . $this->sql_prefix . 'meta;', 1, 0));
   if($row[0] < 1)
   {
    $this->db->insert(/**$table*/$this->sql_prefix . 'meta', array('changes_master'=>0, 'changes_slave'=>0), /**$id_field*/'prikey', /**$execute*/1, /**$stripslashes*/0, /**$insert_key*/0);
   }
  }

  /**
   * Boolscher Check ob uebergebener Typ erlaubt ist
   *
   * @param  string $type
   * @return boolean
   * @access private
  */
  private function is_supported_type($type)
  {
   if(in_array($type, $this->types))
   {
    return true;
   }
   else
   {
    return false;
   }
  }

  /**
   * Liste aller vorhandenen Domains
   * als Array zurueckgeben
   *
   * @param  string $filter  Optionaler Filter fuer den Domain-Namen
   * @return array (id, name)
   * @access public
  */
  public function domain_list($filter='')
  {
   $liste = array();
   $sql = 'select zone_id, zone_name from ' . $this->sql_prefix . 'zones';
   if(!empty($filter))
   {
    $sql .= " where zone_name like '%" . $this->db->addslashes($filter) . "%'";
   }
   $sql .= ' order by zone_name;';
   foreach($this->db->query_fetch_all($sql) as $row)
   {
    extract($row);
    $liste[$zone_id] = $zone_name;
   }
   return $liste;
  }

  /**
   * Domain anlegen
   * Sicherstellen dass diese nicht bereits existiert
   *
   * @param  string  $domain      Domain-Name
   * @return integer Domain-ID oder 0 im Fehlerfall
   * @access public
  */
  public function domain_add($domain)
  {
   /** Whitespaces entfernen */
   $domain = strtolower(trim($domain));
   /** Check auf korrekte Syntax */
   if(!$this->check_domain_name($domain))
   {
    $GLOBALS['cl_api']->log('FAILED: add domain "' . $domain . '" (invalid)', 'dns');
    return 0;
   }
   /** Check auf Existenz */
   if($this->domain_exists($domain))
   {
    $GLOBALS['cl_api']->log('FAILED: add domain "' . $domain . '" (exists)', 'dns');
    return 0;
   }
   /** Abbrechen wenn auf Blacklist */
   if(in_array($domain, $this->blacklist))
   {
    $GLOBALS['cl_api']->log('FAILED: add domain "' . $domain . '" (blacklisted)', 'dns');
    return 0;
   }
   /** Eintrag erstellen */
   $id = $this->db->insert
   (
    /**$table*/$this->sql_prefix . 'zones',
    /**$data*/array
    (
     'zone_name'       => $domain,
     'zone_new_master' => 1,
     'zone_new_slave'  => 1,
     'zone_serial'     => intval(strftime('%Y%m%d') . '01'),
     'zone_spf_strict' => $this->default_spf_strict,
    ),
    /**$id_field*/'zone_id',
    /**$execute*/1,
    /**$stripslashes*/0,
    /**$insert_key*/0
   );
   /** Domain-Datensatz wurde angelegt, Standard-Eintraege anlegen */
   if($id > 0)
   {
    if(!empty($this->default_host_ip))
    {
     $this->entry_add($domain, 'A', '', $this->default_host_ip);
    }
    if(!empty($this->default_host_cname))
    {
     $this->entry_add($domain, 'CNAME', 'www', $this->default_host_cname);
    }
    if(!empty($this->default_spf_new) && is_array($this->default_spf_new))
    {
     foreach($this->default_spf_new as $spf_value)
     {
      $this->entry_add(/**$domain*/$domain, /**$type*/'SPF', /**$name*/'', /**$value*/$spf_value, /**comment*/'');
     }
    }
    /** Zone aktualisieren */
    if(!$this->batch_mode)
    {
     $this->domain_refresh_zones($domain);
    }
    /** Domain-ID zurueckgeben */
    $GLOBALS['cl_api']->log('OK: add domain "' . $domain . '"', 'dns');
    return $id;
   }
   /** Datensatz wurde scheinbar nicht angelegt */
   else
   {
    $GLOBALS['cl_api']->log('FAILED: add domain "' . $domain . '"', 'dns');
    return 0;
   }
  }

  /**
   * Boolsche Abfrage ob Domain existiert
   *
   * @param  string $domain
   * @return boolean
   * @access public
  */
  public function domain_exists($domain)
  {
   $liste = $this->domain_list();
   foreach($liste as $check)
   {
    if($domain == $check)
    {
     return true;
     break;
    }
   }
  }

  /**
   * Komplette Domain inkl. aller Entrys loeschen
   * Die Zone-Files werden am Server NICHT automatisch entfernt/ausgetragen
   * Dies muss manuell erfolge, sie werden aber nicht mehr neu erstellt
   *
   * @param  string $domain
   * @return boolean
   * @access public
  */
  public function domain_remove($domain)
  {
   $zone_id = $this->domain_get_id($domain);
   if($zone_id > 0)
   {
    /** Alle verknuepften Entrys entfernen */
    $this->db->query('delete from ' . $this->sql_prefix . 'entrys where entry_zone_id=' . (int)$zone_id . ';', 1, 1);
    /** Zone selbst entfernen */
    $this->db->query('delete from ' . $this->sql_prefix . 'zones where zone_id=' . (int)$zone_id . ';', 1, 1);
    /** Erfolg zurueckgeben */
    $GLOBALS['cl_api']->log('OK: remove domain "' . $domain . '"', 'dns');
    return true;
   }
   /** Uebergebene Domain scheint nicht zu existieren */
   else
   {
    $GLOBALS['cl_api']->log('FAILED: remove domain "' . $domain . '"', 'dns');
    return false;
   }
  }

  /**
   * Eintrag zur angegebenen Domain hinzufuegen
   *
   * @param  string $domain   Domain
   * @param  string $type     A, CNAME, MX oder SPF
   * @param  string $name     Name fuer den Eintrag (www, mail etc. -> Ohne Angabe wird dies im Fall eines A-Entrys der Domain-Eintrag)
   * @param  string $value    Wert
   * @param  string $comment  Interner Kommentar
   * @return boolean
   * @access public
  */
  public function entry_add($domain, $type, $name, $value, $comment='')
  {
   /** Whitespaces filtern */
   $domain  = strtolower(trim($domain));
   $type    = strtoupper(trim($type));
   $comment = trim($comment);
   switch($type)
   {
    case 'TXT':
      $name  = trim($name);
      $value = trim($value);
      break;
    default:
      $name  = strtolower(trim($name));
      $value = strtolower(trim($value));
      break;
   }
   /** Sicherstellen dass nur unterstuetzte Typen eingetragen werden */
   if(!in_array($type, $this->types) || empty($value) || empty($type) || empty($domain))
   {
    $GLOBALS['cl_api']->log('FAILED: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
    return false;
   }
   /** Sicherstellen dass bei CNAMES Name und Wert uebergeben wurde */
   if($type == 'CNAME' && $name == '')
   {
    $GLOBALS['cl_api']->log('FAILED: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
    return false;
   }
   /** Sicherstellen dass Domain in der Datenbank existiert */
   $domain_id = $this->domain_get_id($domain);
   if(!$domain_id)
   {
    $GLOBALS['cl_api']->log('FAILED: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
    return false;
   }
   /** Wenn A-Entry und kein Name alle A-Entrys der Domain ohne Namen im Vorfeld entfernen */
   if($type == 'A' && $name == '')
   {
    $this->db->query("delete from " . $this->sql_prefix . "entrys where entry_name='' and entry_type='A' and entry_zone_id=" . (int)$domain_id . ';', 1, 1);
   }
   /** CNAME/MX/NS-Eintraege die mehr als einen Punkt enthalten und nicht mit einem solchen abschliessen automatisch ergaenzen */
   if($type == 'CNAME' || $type == 'MX' || $type == 'NS')
   {
    $value = $this->check_cname_mx($value);
   }
   /** A-Entrys deren Name mehr als einen Punkt enthaelt und nicht mit einem solchen abschliessen automatisch ergaenzen */
   if($type == 'A')
   {
    $name = $this->check_a_name($name);
   }
   /** Sicherstellen dass keine doppelten Eintraege angelegt werden */
   switch($type)
   {
    /* Value muss einmalig sein */
    case 'SPF':
      $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where entry_type='SPF' and entry_zone_id=" . $domain_id . " and entry_value='" . $this->db->addslashes($value) . "';", 1, 0));
      if($row[0] > 0)
      {
       $GLOBALS['cl_api']->log('FAILED: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
       return false;
      }
      break;
    /** Value muss einmalig sein und darf nicht mit einem CNAME kollidieren */
    case 'MX':
      $mx_exists = 0;
      /** Einmaligkeit sicherstellen */
      $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where entry_type='MX' and entry_zone_id=" . $domain_id . " and entry_name='" . $this->db->addslashes($name) .  "' and entry_value='" . $this->db->addslashes($value) . "';", 1, 0));
      if($row[0] > 0)
      {
       $mx_exists = 1;
      }
      /** Kollision mit CNAME abfangen */
      $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where entry_type='CNAME' and entry_zone_id=" . $domain_id . " and entry_name='" . $this->db->addslashes($name) .  "';", 1, 0));
      if($row[0] > 0)
      {
       $mx_exists = 1;
      }
      /** Abbrechen */
      if($mx_exists)
      {
       $GLOBALS['cl_api']->log('FAILED: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
       return false;
      }
      break;
    /** Value muss einmalig sein und darf nicht mit einem CNAME kollidieren */
    case 'TXT':
      $txt_exists = 0;
      /** Einmaligkeit sicherstellen */
      $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where entry_type='TXT' and entry_zone_id=" . $domain_id . " and entry_name='" . $this->db->addslashes($name) .  "' and entry_value='" . $this->db->addslashes($value) . "';", 1, 0));
      if($row[0] > 0)
      {
       $txt_exists = 1;
      }
      /** Kollision mit CNAME abfangen */
      $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where entry_type='CNAME' and entry_zone_id=" . $domain_id . " and entry_name='" . $this->db->addslashes($name) .  "';", 1, 0));
      if($row[0] > 0)
      {
       $txt_exists = 1;
      }
      /** Abbrechen */
      if($txt_exists)
      {
       $GLOBALS['cl_api']->log('FAILED: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
       return false;
      }
      break;
    /** Value muss einmalig sein und darf nicht mit einem CNAME kollidieren */
    case 'NS':
      $ns_exists = 0;
      /** Einmaligkeit sicherstellen */
      $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where entry_type='NS' and entry_zone_id=" . $domain_id . " and entry_name='" . $this->db->addslashes($name) .  "' and entry_value='" . $this->db->addslashes($value) . "';", 1, 0));
      if($row[0] > 0)
      {
       $ns_exists = 1;
      }
      /** Kollision mit CNAME abfangen */
      $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where entry_type='CNAME' and entry_zone_id=" . $domain_id . " and entry_name='" . $this->db->addslashes($name) .  "';", 1, 0));
      if($row[0] > 0)
      {
       $ns_exists = 1;
      }
      /** Abbrechen */
      if($ns_exists)
      {
       $GLOBALS['cl_api']->log('FAILED: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
       return false;
      }
      break;
    /** Name muss einmalig sein und darf auch mit keinem CNAME kollidieren */
    case 'A':
      $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where (entry_type='A' or entry_type='CNAME') and entry_zone_id=" . $domain_id . " and entry_name='" . $this->db->addslashes($name) . "';", 1, 0));
      if($row[0] > 0)
      {
       $GLOBALS['cl_api']->log('FAILED: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
       return false;
      }
      break;
    /** Name muss einmalig sein und darf auch mit keinem A/CNAME/MX kollidieren */
    case 'CNAME':
      $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where (entry_type='A' or entry_type='CNAME' or entry_type='MX' or entry_type='TXT') and entry_zone_id=" . $domain_id . " and entry_name='" . $this->db->addslashes($name) . "';", 1, 0));
      if($row[0] > 0)
      {
       $GLOBALS['cl_api']->log('FAILED: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
       return false;
      }
      break;
   }
   /** Eintrag hinzufuegen */
   $insertid = $this->db->insert
   (
    /**$table*/$this->sql_prefix . 'entrys',
    /**$data*/array
    (
     'entry_zone_id'        => $domain_id,
     'entry_zone_cleartext' => $domain,
     'entry_type'           => $type,
     'entry_name'           => $name,
     'entry_value'          => $value,
     'entry_comment'        => $comment,
     'entry_mac'            => '',
    ),
    /**$id_field*/'entry_id',
    /**$execute*/1,
    /**$stripslashes*/0,
    /**$insert_key*/0
   );
   /** Erfolg zurueckgeben */
   if($insertid)
   {
    /** Zonen aktualisieren */
    if(!$this->batch_mode)
    {
     $this->domain_refresh_zones($domain);
    }
    /** Erfolg zurueckgeben */
    $GLOBALS['cl_api']->log('OK: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
    return true;
   }
   /** SQL-Insert war nicht erfolgreich */
   else
   {
    $GLOBALS['cl_api']->log('FAILED: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
    return false;
   }
  }

  /**
   * Boolsche Abfrage ob in der angegebenen Domain ein
   * bestimmter Eintrag existiert
   *
   * @param  string $domain
   * @param  string $type
   * @param  string $name
   * @return boolean
   * @access public
  */
  public function entry_exists($domain, $type, $name='')
  {
   /** Sicherstellen dass Domain in der Datenbank existiert */
   $domain_id = $this->domain_get_id($domain);
   if($domain_id < 1)
   {
    return false;
   }
   else
   {
    $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where entry_type='" . $this->db->addslashes($type) . "' and entry_zone_id=" . $domain_id . " and entry_name='" . $this->db->addslashes($name) . "';", 1, 0));
    if($row[0] > 0)
    {
     return true;
    }
    else
    {
     return false;
    }
   }
  }

  /**
   * Boolsche Abfrage ob in der angegebenen Domain ein bestimmter Eintrag
   * unabhaengig vom Typ existiert um beispielsweise CNAMES automatisiert
   * anzulegen wenn diese nicht vorhanden sind und dabei ggf. vorhandene
   * A-Recors mit dem gleichen Namen sauber zu berucksichtigen
   *
   * @param  string $domain
   * @param  string $name
   * @access public
   * @return boolean
  */
  public function entry_exists_name($domain, $name)
  {
   /** Sicherstellen dass Domain in der Datenbank existiert */
   $domain_id = $this->domain_get_id($domain);
   if($domain_id < 1)
   {
    return false;
   }
   else
   {
    $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "entrys where entry_zone_id=" . $domain_id . " and entry_name='" . $this->db->addslashes($name) . "';", 1, 0));
    if($row[0] > 0)
    {
     return true;
    }
    else
    {
     return false;
    }
   }
  }

  /**
   * Datensatz aus der Tabelle "entrys" entfernen
   *
   * @param  integer $key  Datensatz-ID
   * @return boolean
   * @access public
  */
  public function entry_remove($key)
  {
   /** Domain ermitteln */
   $data = $this->entry_get($key);
   if(!empty($data))
   {
    /** Eintrag entfernen */
    if($this->db->query('delete from ' . $this->sql_prefix . 'entrys where entry_id=' . (int)$key) . ';')
    {
     /** Zonen aktualisieren */
     if(!$this->batch_mode)
     {
      $this->domain_refresh_zones($data['entry_zone_cleartext']);
     }
     $GLOBALS['cl_api']->log('OK: "' . $data['entry_zone_cleartext'] . '": Remove ' . $data['entry_type'] . ', ' . "'" . $data['entry_name'] . "' => "  . $data['entry_value'], 'dns');
     return true;
    }
    /** SQL-Abfrage fehlgeschlagen */
    else
    {
     $GLOBALS['cl_api']->log('FAILED: "' . $data['entry_zone_cleartext'] . '": Remove ' . $data['entry_type'] . ', ' . "'" . $data['entry_name'] . "' => "  . $data['entry_value'], 'dns');
     return false;
    }
   }
   else
   {
    $GLOBALS['cl_api']->log('FAILED: Remove entry "' . $key . '"', 'dns');
    return false;
   }

  }

  /**
   * Eintrag bearbeiten
   *
   * @param  integer $key     entry_id
   * @param  integer $value   entry_value
   * @param  string  $comment entry_comment
   * @return boolean
   * @access public
  */
  public function entry_edit_value($key, $value, $comment='')
  {
   /** Datensatz auslesen */
   $key = (int)$key;
   $data = $this->entry_get($key);
   if(empty($data))
   {
    return false;
   }
   switch($data['entry_type'])
   {
    case 'TXT': $value = trim($value); break;
    default:    $value = strtolower(trim($value)); break;
   }
   /** Zwischenspeichern der aktuellen Metadaten */
   $domain  = $data['entry_zone_cleartext'];
   $zone_id = $data['entry_zone_id'];
   $type    = $data['entry_type'];
   $name    = $data['entry_name'];
   /** CNAME/MX-Eintraege die mehr als einen Punkt enthalten und nicht mit einem solchen abschliessen automatisch ergaenzen */
   if($type == 'CNAME' || $type == 'MX')
   {
    $value = $this->check_cname_mx($value);
   }
   /** Aenderung durchfuehren */
   $this->db->query("update " . $this->sql_prefix . "entrys set entry_value='" . $this->db->addslashes($value) . "', entry_comment='" . $this->db->addslashes($comment) . "' where entry_id=" . (int)$key . ';', 1, 0);
   /** Bei MX und SPF sicherstellen dass es fuer eine Domain keine doppelten Eintraege gibt */
   if($type == 'MX')
   {
    $this->db->query('delete from ' . $this->sql_prefix . 'entrys where entry_zone_id=' . (int)$zone_id . " and entry_name='" . $this->db->addslashes($name) .  "' and entry_type='" . $this->db->addslashes($type) . "' and entry_value='" . $this->db->addslashes($value) . "' and entry_id!=" . (int)$key . ';', 1, 1);
   }
   elseif($type == 'TXT')
   {
    $this->db->query('delete from ' . $this->sql_prefix . 'entrys where entry_zone_id=' . (int)$zone_id . " and entry_name='" . $this->db->addslashes($name) .  "' and entry_type='" . $this->db->addslashes($type) . "' and entry_value='" . $this->db->addslashes($value) . "' and entry_id!=" . (int)$key . ';', 1, 1);
   }
   elseif($type == 'SPF')
   {
    $this->db->query('delete from ' . $this->sql_prefix . 'entrys where entry_zone_id=' . $zone_id . " and entry_type='" . $this->db->addslashes($type) . "' and entry_value='" . $this->db->addslashes($value) . "' and entry_id!=" . (int)$key . ';', 1, 1);
   }
   /** Zonen aktualisieren */
   if(!$this->batch_mode)
   {
    $this->domain_refresh_zones($domain);
   }
   $GLOBALS['cl_api']->log('OK: "' . $domain . '": ' . $type . ', ' . "'" . $name . "' => "  . $value, 'dns');
   return true;
  }

  /**
   * MAC-Adresse eines A-Entry setzen
   * @param  integer $entry_id
   * @param  string  $value
   * @access public
   * @return boolean
  */
  public function entry_set_mac($entry_id, $value)
  {
   /** White-Spaces filtern */
   $value = str_replace('-', ':', strtolower(trim($value)));
   /** Sicherstellen dass wir einen A-Record bearbeiten */
   $old = $this->entry_get($entry_id);
   if($old['entry_type'] != 'A' || (strlen($value) != 17 && !empty($value)))
   {
    return false;
   }
   /** Entfernen ohne weitere Checks */
   if(empty($value))
   {
    $this->db->query('update ' . $this->sql_prefix . 'entrys set entry_mac=\'\' where entry_id=' . (int)$entry_id . ';', 1, 1);
    return true;
   }
   /** Sicherstellen dass nicht bereits ein anderer Eintrag diese MAC zugewiesen hat */
   else
   {
    /** Existenz checken */
    $row = $this->db->fetch_row($this->db->query('select count(*) from ' . $this->sql_prefix . 'entrys where entry_id!=' . (int)$entry_id . " and entry_mac='" . $this->db->addslashes($value) . "';", 1, 0));
    if($row[0] > 0)
    {
     return false;
    }
    /** Zuweisen */
    else
    {
     $this->db->query('update ' . $this->sql_prefix . 'entrys set entry_mac=\'' . $this->db->addslashes($value) . '\' where entry_id=' . (int)$entry_id . ';', 1, 1);
     return true;
    }
   }
  }

  /**
   * Abfrage ob ein angegebener Eintrag ein A-Record ist
   *
   * @param  string $domain
   * @param  string $name
   * @return boolean
   * @access public
  */
  public function is_a_record($domain, $name)
  {
   $row = $this->db->fetch_row($this->db->query('select count(*) from ' . $this->sql_prefix. "entrys where entry_type='A' and entry_zone_cleartext='" . $this->db->addslashes($domain). "' and entry_name='" . $this->db->addslashes($name) . "';", 1, 0));
   if($row[0] > 0)
   {
    return true;
   }
   else
   {
    return false;
   }
  }

  /**
   * Eintrag als Array auslesen
   *
   * @param  integer $key
   * @return array
   * @access public
  */
  public function entry_get($key)
  {
   return $this->db->query_record(/**$table*/$this->sql_prefix . 'entrys', /**$key_field*/'entry_id', /**$key*/$key, /**$mssql_trim*/0);
  }

  /**
   * Zone-ID anhand des Domain-Namens ermitteln
   *
   * @param  string $domain
   * @return integer
   * @access public
  */
  public function domain_get_id($domain)
  {
   $domain = trim($domain);
   $result = $this->db->query("select zone_id from " . $this->sql_prefix . "zones where zone_name='" . $this->db->addslashes($domain) . "';", 1, 0);
   if($this->db->num_rows($result) < 1)
   {
    return 0;
   }
   else
   {
    $row = $this->db->fetch_row($result);
    return $row[0];
   }
  }

  /**
   * Interne Zone-ID anhand einer Full-Qualified-Eingabe ermitteln
   *
   * @param  string $fq Einageb eines kompletten Hosts inkl. Domain
   * @return integer
   * @access public
  */
  public function domain_get_id_by_fq($fq)
  {
   /** Sicherstellen dass wir eine saubere Eingabe haben */
   $fq = $this->normalize_fq($fq);
   /** Domain ermitteln */
   $parts = explode('.', $fq);
   $count = count($parts) - 1;
   $zsp = '';
   for($i=$count; $i>=0; $i--)
   {
    $zsp = $parts[$i] . '.' . $zsp;
    $zsp2 = $zsp;
    $zsp2_len = strlen($zsp2);
    if(substr($zsp2, ($zsp2_len-1), 1) == '.')
    {
     $zsp2 = substr($zsp2, 0, ($zsp2_len-1));
    }
    if($this->domain_exists($zsp2))
    {
     return (int)$this->domain_get_id($zsp2);
     break;
    }
   }
   /** Wenn kein Eintrag gefunden wurde 0 liefern */
   return 0;
  }

  /**
   * Boolsche Abfrage ob die angegebene Domain MX-Eintraege hat
   *
   * Wird primaer beim Generieren von SPF-Records
   * benutzt um dynamisch "v=spf1 -all" zu setzen
   *
   * @param  string $domain
   * @return boolean
   * @access public
  */
  public function domain_has_mx($domain)
  {
   $row = $this->db->fetch_row($this->db->query('select count(*) from ' . $this->sql_prefix . "entrys where entry_zone_cleartext='" . $this->db->addslashes($domain) . "' and entry_type='MX';", 1, 0));
   if($row[0] > 0)
   {
    return true;
   }
   else
   {
    return false;
   }
  }

  /**
   * Abschliessenden Punkt bei Gost-Namen entfernen
   *
   * @param  string $fq
   * @return string
   * @access private
  */
  private function normalize_fq($fq)
  {
   $fq = trim($fq);
   $fq_len = strlen($fq);
   if(substr($fq, ($fq_len-1), 1) == '.')
   {
    $fq = substr($fq, 0, ($fq_len-1));
   }
   return $fq;
  }

  /**
   * IP-Adresse eines Full-Qualified-Entries ermitteln
   *
   * Verwendet primaer die Datenbank, kann aber als
   * Fallback auch eine DNS-Abfrage starten
   *
   * @param  string  $fq          FQ Hostname
   * @param  boolean  $allow_dns  Wenn kein Eintrag in Datenbank DNS-Request erlauben
   * @return string
   * @access public
  */
  public function find_address($fq, $allow_dns=false)
  {
   /** Sicherstellen dass wir eine saubere Eingabe haben */
   $fq = $this->normalize_fq($fq);
   /** Statischer Array um die Performance zu verbessern da diese Methode beim Generiren von SPF-Records vielfach zur Anwendung kommt */
   static $address_buffer;
   if(isset($address_buffer[$fq]))
   {
    return $address_buffer[$fq];
   }
   /** Zone-ID ermitteln */
   $zone_id = $this->domain_get_id_by_fq($fq);
   /** Wenn Zone-ID in der Datenbank existiert interne Weiterverarbeitung */
   if($zone_id > 0)
   {
    /** Cleartext-Domain und gesuchten Name zwischenspeichern */
    $domain = $this->domain_get_name_by_id($zone_id);
    $entry_name = str_replace('.' . $domain, '', $fq);
    /** SQL-Abfrage nach dem gesuchten Eintrag */
    $result = $this->db->query('select * from ' . $this->sql_prefix . 'entrys where entry_zone_id=' . (int)$zone_id . " and (entry_type='A' or entry_type='CNAME') and (entry_name='" . $this->db->addslashes($entry_name) . "' or entry_name='" . $this->db->addslashes($fq . '.') . "');", 1, 0);
    if($this->db->num_rows($result))
    {
     /** Result in Array zwischenspeichern */
     $row = $this->db->fetch_assoc($result);
     $row['entry_value'] = trim($row['entry_value']);
     /** Ergebnis eines A-Records direkt zurueckgeben */
     if($row['entry_type'] == 'A')
     {
      $address_buffer[$fq] = $row['entry_value'];
      return $row['entry_value'];
     }
     /** CNAME aufloesen */
     elseif($row['entry_type'] == 'CNAME')
     {
      /** Wenn moeglich A-Record in der eigenen Domain */
      $result = $this->db->query('select * from ' . $this->sql_prefix . 'entrys where entry_zone_id=' . (int)$zone_id . " and (entry_type='A' or entry_type='CNAME') and (entry_name='" . $this->db->addslashes($row['entry_value']) . "' or entry_name='" . $this->db->addslashes($fq . '.') . "');", 1, 0);
      if($this->db->num_rows($result))
      {
       /** Result in Array zwischenspeichern */
       $row = $this->db->fetch_assoc($result);
       $row['entry_value'] = trim($row['entry_value']);
       /** Ergebnis eines A-Records direkt zurueckgeben */
       if($row['entry_type'] == 'A')
       {
        $address_buffer[$fq] = $row['entry_value'];
        return $row['entry_value'];
       }
       /** CNAME rekursiv aufloesen */
       else
       {
        $rw = $this->find_address($row['entry_value'], /**$allow_dns*/false);
        $address_buffer[$fq] = $rw;
        return $rw;
       }
      }
      /** Rekursiver Call wenn CNAME in fremde Domain zeigt*/
      else
      {
       $rw = $this->find_address($row['entry_value'], /**$allow_dns*/false);
       $address_buffer[$fq] = $rw;
       return $rw;
      }
     }
    }
   }
   /** Wenn aktiv als Fallback DNS-Abfrage verwenden */
   elseif($allow_dns)
   {
    $ip = gethostbyname($fq);
    if($ip != $fq)
    {
     $address_buffer[$fq] = $ip;
     return $ip;
    }
   }
   /** Wenn kein Eintrag gefunden wurde einen Leerstring liefern */
   $address_buffer[$fq] = '';
   return '';
  }

  /**
   * Domain-Namen anhand der Zone-ID abfragen
   *
   * @param  integer $id Zone-ID
   * @return string
   * @access public
  */
  public function domain_get_name_by_id($id)
  {
   $result = $this->db->query("select zone_name from " . $this->sql_prefix . "zones where zone_id=" . (int)$id . ';', 1, 0);
   if($this->db->num_rows($result) < 1)
   {
    return '';
   }
   else
   {
    $row = $this->db->fetch_row($result);
    return $row[0];
   }
  }

  /**
   * Zone-Eintraege in den Haupt-Datensaetzen neu generieren
   *
   * Erstellt im Prinzip saemtliche Domains neu mit eindeutigen Seriennummern
   * Zum Einen verhindert dies eine Serial-Kollision
   *
   * Zum Anderen koennen globale Settings (Expires, Default-SPF..) automatisch auf
   * alle Domains angwendet werden
   *
   * @param  $domain  Optionale Angabe einer einzelnen Domain
   * @return void
   * @access public
  */
  public function domain_refresh_zones($domain='')
  {
   /** Wenn Domain uebergeben ID abfragen */
   if(!empty($domain))
   {
    $liste = array();
    $result = $this->db->query_fetch_all("select zone_id, zone_name from " . $this->sql_prefix . "zones where zone_name like '" . $this->db->addslashes($domain) . "';");
    if(count($result) > 0)
    {
     foreach($result as $row)
     {
      $zsp_id = $row['zone_id'];
      $liste[$zsp_id] = $row['zone_name'];
     }
     $GLOBALS['cl_api']->log('Refresh "' . $domain . '"', 'dns');
    }
   }
   /** Alle Domains aktualisieren */
   else
   {
    /** Log */
    $GLOBALS['cl_api']->log('Refresh all Zone-Files', 'dns');
    /** Domain-Liste abrufen */
    $liste = $this->domain_list();
   }
   /** Durchlaufen */
   if(!empty($liste))
   {
    /** Domains durchlaufen */
    foreach($liste as $id=>$name)
    {
     /** Leeres Zone-File fuer durchlaufende Domain starten */
     $data = '';
     /** Seriennummer fuer jede Domain 1 hochzaehlen */
     $serial = $this->next_serial($id);
     /** Default TTL */
     $data .= '$TTL ' . $this->domain_get_ttl($name) . MY_LE . MY_LE;
     /** SOA und tech. Admin */
     $data .= '@ IN SOA ' . $this->soa . '. ' . str_replace('@', '.', $this->hostmaster) . '. (' . MY_LE;
     /** Eindeutige Serial */
     $data .= "\t\t\t\t" . $serial . "\t" . '; Serial' . MY_LE;
     /** Aktualisierungs-Interval */
     $data .= "\t\t\t\t" . $this->refresh . "\t\t" . '; Refresh' . MY_LE;
     /** Interval fuer erneute Versuche bei Transfer-Fehlern */
     $data .= "\t\t\t\t" . $this->retry . "\t\t" . '; Retry' . MY_LE;
     /** Ablauf der Zone wenn Master lange Zeit nicht verfuegbar ist */
     $data .= "\t\t\t\t" . $this->expire . "\t\t" . '; Expire' . MY_LE;
     /** TTL fuer negative Antworten */
     $data .= "\t\t\t\t" . $this->negative_ttl . "\t\t" . '; Negative-TTL' . MY_LE;
     /** Rahmendaten abschliessen */
     $data .= ');' . MY_LE;
     /**
      localhost-Eintrag laut https://tools.ietf.org/html/rfc1537
      Note that all domains that contain hosts should have a
      "localhost" A record in them
     */
     $data .= MY_LE . 'localhost' . "\t\t\t" . 'IN A' . "\t\t127.0.0.1" . MY_LE;
     /** Nameserver-Eintrag */
     $data .= MY_LE . '@' . "\t\t\t\t" . 'IN NS' . "\t\t" . $this->soa . '.' . MY_LE;
     /** Zweiter Nameserver-Eintrag  wenn konfiguriert */
     if(!empty($this->ns2))
     {
      $data .= '@' . "\t\t\t\t" . 'IN NS' . "\t\t" . $this->ns2 . '.' . MY_LE;
     }
     /** Dritter Nameserver-Eintrag  wenn konfiguriert */
     if(!empty($this->ns3))
     {
      $data .= '@' . "\t\t\t\t" . 'IN NS' . "\t\t" . $this->ns3 . '.' . MY_LE;
     }
     $data .= MY_LE;
     /** Default-A-Host wenn eingetragen */
     $data .= $this->domain_refresh_default_a($id);
     /** Sonstige A-Hosts */
     $data .= $this->domain_refresh_a($id);
     /** MX */
     $data .= $this->domain_refresh_mx($id);
     /** SPF-Eintrag */
     $data .= $this->domain_refresh_spf($id, /**$record_type*/'TXT');
     $data .= $this->domain_refresh_spf($id, /**$record_type*/'SPF');
     /** Erweiterte TXT-Records */
     $data .= $this->domain_refresh_txt($id);
     /** DNS-Delegation */
     $data .= $this->domain_refresh_ns($id);
     /** CNAMES */
     $data .= $this->domain_refresh_cname($id);
     /** Generiertes Zone-File in Datenbank ablegen */
     $this->db->query("update " . $this->sql_prefix . "zones set zone_serial='" . $serial . "',zone_content='" . $this->db->addslashes($data) . "' where zone_id=" . (int)$id . ';', 1, 1);
     /** Wenn LAN/WAN-Translation konfiguriert Daten fuer internen DNS generieren und ablegen */
     if(!empty($this->lan_translation))
     {
      $data_lan = $data;
      foreach($this->lan_translation as $translation)
      {
       $data_lan = str_replace($translation['source'], $translation['target'], $data_lan);
      }
      $this->db->query("update " . $this->sql_prefix . "zones set zone_content_lan='" . $this->db->addslashes($data_lan) . "' where zone_id=" . (int)$id . ';', 1, 1);
     }
    }
    unset($liste);
    $this->notify_changes();
   }
   /** SQL-Tabellen optimieren */
   $this->db->query('optimize table ' . $this->sql_prefix . 'zones;', 1, 1);
   $this->db->query('optimize table ' . $this->sql_prefix . 'entrys;', 1, 1);
  }

  /**
   * Naechste frei Zone-Serial abfragen
   *
   * Wenn moeglich aktuellen Timestamp, ansonsten hoechster benutzter Wert + 1
   * Fallbacks um parallele Aktualisierungen durch entsprechende Erhoehungen abzufangen
   *
   * @param  integer $zone_id
   * @return integer
   * @access public
  */
  public function next_serial($zone_id)
  {
   /** Versuchen aktuellen Timestamp zu benutzen wenn nicht bereits vergeben */
   $next_serial = intval(strftime('%Y%m%d') . '01');
   /** Check ob Serial bereits in Verwendung ist */
   $count_row = $this->db->fetch_row($this->db->query('select count(*) from ' . $this->sql_prefix . 'zones where zone_id=' . (int)$zone_id . ' and zone_serial>=' . (int)$next_serial . ';', 1, 0));
   if($count_row[0] > 0)
   {
    /** Aktuell hoechsten Wert um 1 erhoeht abfragen */
    $row = $this->db->fetch_row($this->db->query('select max(zone_serial)+1 from ' . $this->sql_prefix . 'zones where zone_id=' . (int)$zone_id . ';', 1, 0));
    $next_serial = $row[0];
    /** Check ob Serial bereits in Verwendung ist */
    $count_row = $this->db->fetch_row($this->db->query('select count(*) from ' . $this->sql_prefix . 'zones where zone_id=' . (int)$zone_id . ' and zone_serial>=' . (int)$next_serial . ';', 1, 0));
    if($count_row[0] > 0)
    {
     /** Aktuell hoechsten Wert um 1 erhoeht abfragen */
     $row = $this->db->fetch_row($this->db->query('select max(zone_serial)+1 from ' . $this->sql_prefix . 'zones where zone_id=' . (int)$zone_id . ';', 1, 0));
     $next_serial = $row[0];
    }
   }
   /** Neue nicht vergebene Serial zurueckgeben */
   unset($row, $count_row);
   return $next_serial;
  }

  /** Changes in Tabelle "meta" notieren */
  private function notify_changes()
  {
   $this->db->query('update ' . $this->sql_prefix . 'meta set changes_master=1, changes_slave=1;', 1, 1);
  }

  /**
   * MX-Eintraege fuer die uebergebene DNS-Zone generieren
   *
   * @param integer $zone_id  Domain-ID
   * @return string
   * @access private
  */
  private function domain_refresh_mx($zone_id)
  {
   $data        = '';
   $mx_priority = array();
   /** Alle Eintraege vom Typ "MX" der angegebenen Domain sortiert nach Zeit des Anlegens abfragen */
   $result = $this->db->query_fetch_all("select entry_name, entry_value from " . $this->sql_prefix . "entrys where entry_type='MX' and entry_zone_id=" . (int)$zone_id. ' order by entry_id;');
   if(count($result) > 0)
   {
    $data .= MY_LE;
    foreach($result as $row)
    {
     /** Sicherstellen dass wir einen Zonen-Namen verarbeiten */
     if(empty($row['entry_name']))
     {
      $row['entry_name'] = '@';
     }
     /** Priority hochzaehlen */
     if(empty($row['entry_name']))
     {
      $pri_base = 'base';
     }
     else
     {
      $pri_base = $row['entry_name'];
     }
     settype($mx_priority[$pri_base], 'integer');
     $mx_priority[$pri_base] += 10;
     /** Eintrag generieren */
     $data .= $row['entry_name'] . "\t\t\t";
     if(strlen($row['entry_name']) <= $this->tab_fuzz)
     {
      $data .= "\t";
     }
     $data .= 'IN MX' . "\t\t";
     $data .= $mx_priority[$pri_base] . ' ' . $row['entry_value'];
     $data .= MY_LE;
    }
    unset($result);
   }
   return $data;
  }

  /**
   * Default-A-Eintrag fuer die Domain ohne Hostname generieren (zB: thelounge.net direkt eingegeben)
   *
   * @param  integer $zone_id  Domain-ID
   * @return string
   * @access private
  */
  private function domain_refresh_default_a($zone_id)
  {
   $data = '';
   /** Alle Eintraege vom Typ "A" die keinen Namen haben abfragen, pro Domain einmal moeglich */
   $result = $this->db->query("select entry_value from " . $this->sql_prefix . "entrys where entry_name='' and entry_type='A' and entry_zone_id=" . (int)$zone_id . ';', 1, 0);
   if($this->db->num_rows($result))
   {
    $row = $this->db->fetch_row($result);
    $row[0] = $this->get_lounge_ip($row[0]);
    $data .= '@' . "\t\t\t\t" . 'IN A' . "\t\t" . trim($row[0]);
    $data .= MY_LE;
   }
   return $data;
  }

  /**
   * A-Host-Eintraege der Domain generieren
   *
   * @param  integer $zone_id  Domain-ID
   * @return string
   * @access private
  */
  private function domain_refresh_a($zone_id)
  {
   $data = '';
   /** Alle Eintraege vom Typ "A" die einen Namen vergeben haben sortiert nach diesem abfragen */
   $result = $this->db->query_fetch_all("select entry_name, entry_value from " . $this->sql_prefix . "entrys where entry_name!='' and entry_type='A' and entry_zone_id=" . (int)$zone_id . ' order by entry_name;');
   if(count($result) > 0)
   {
    foreach($result as $row)
    {
     $row['entry_value'] = $this->get_lounge_ip($row['entry_value']);
     $data .= $row['entry_name'] . "\t\t";
     if(strlen($row['entry_name']) < 16)
     {
      $data .= "\t";
     }
     if(strlen($row['entry_name']) <= $this->tab_fuzz)
     {
      $data .= "\t";
     }
     $data .= 'IN A' . "\t\t";
     $data .= $row['entry_value'];
     $data .= MY_LE;
    }
    unset($result);
   }
   return $data;
  }

  /**
   * SPF-Eintrag fuer die Domain generieren
   *
   * @param  integer $zone_id     Domain-ID
   * @param  string  $record_type 'TXT' oder 'SPF' um beide Record-Typen zu generieren
   * @return string
   * @access private
  */
  private function domain_refresh_spf($zone_id, $record_type='TXT')
  {
   /** Domain-Name ermitteln */
   $data  = '';
   $domain = $this->domain_get_name_by_id($zone_id);
   /** Wenn die Domain keine MX-Eintraege besitzt No-Mail-SPF generieren */
   if(!$this->domain_has_mx($domain))
   {
    return MY_LE . $data .= '@' . "\t\t\t\t" . 'IN ' . $record_type . "\t\t" . '"v=spf1 -all"' . MY_LE;
   }
   /** Alle Eintraege vom Typ "SPF" sortiert nach der Datensatz-Nummer abfragen */
   $result = $this->db->query_fetch_all("select entry_value from " . $this->sql_prefix . "entrys where entry_type='SPF' and entry_zone_id=" . (int)$zone_id . ' order by entry_id;', /**$numeric*/true);
   if(count($result) > 0)
   {
    /** Temporaere Liste initialisieren */
    $spf_liste = array();
    foreach($result as $row)
    {
     /** Sicherstellen dass keine Whitespaces vorkommen */
     $row[0] = trim($row[0]);
     /** Lounge-Standard ist aktiv - Mit globaler Konfiguration abgleichen */
     if($row[0] == 'default')
     {
      foreach($this->default_spf as $akt)
      {
       $spf_liste[] = trim($akt);
      }
     }
     /** Zusaetzliche Eintraege */
     elseif(!in_array($row[0], $spf_liste))
     {
      $spf_liste[] = $row[0];
     }
    }
    /** Wenn SPF-Liste gesetzt ist Eintrag aus allen vorhandenen Daten generieren */
    if(!empty($spf_liste))
    {
     $data .= MY_LE;
     $data .= '@' . "\t\t\t\t";
     /** SPF-Wert fuer eventuelle Subdomains in eigener Variable generieren */
     $spf_value  = 'IN ' . $record_type . "\t\t" . '"v=spf1';
     $spf_string = '';
     foreach($spf_liste as $akt_spf)
     {
      $akt_spf = trim($akt_spf);
      /** Wenn Doppelpunkt im Value steht unveraendert uebernehmen */
      if(stristr($akt_spf, ':'))
      {
       $spf_string .= ' ' . $akt_spf;
      }
      /** ip4-mechansism priorisieren */
      else
      {
       /** CIDR-Angabe eines Netzwerkes als ip4 priorisieren */
       if(strpos($akt_spf, '/') !== false)
       {
        $spf_string .= ' ip4:' . $akt_spf;
       }
       else
       {
        /** Versuchen eine IPv4-Adrese zu ermitteln */
        $ip4 = $this->find_address($akt_spf, /**$allow_dns*/1);
        if(empty($ip4))
        {
         $ip4 = $this->find_address($akt_spf . '.' . $domain, /**$allow_dns*/1);
        }
        if(!empty($ip4))
        {
         $spf_string .= ' ip4:' . $ip4;
        }
        /** Fallback auf a-mechanism */
        else
        {
         $spf_string .= ' a:' . $akt_spf;
        }
       }
      }
     }
     /** SPF-Testing bzw. Scoring-Modus */
     if(!$this->domain_get_spf_strict($domain))
     {
      $spf_string .= ' ~all"';
     }
     /** Strikte SPF-Behandlung aktivieren */
     else
     {
      $spf_string .= ' -all"';
     }
     $spf_value .= $spf_string;
     /** SPF-Wert an Haupteintrag anhaengen */
     $data .= $spf_value . MY_LE;
    }
    /** SPF-Records fuer Subdomains generieren - HACK: Nur wenn MX-Target auf thelounge.net verweist um Subdomains die extern gehostet werden hier zu ignorieren */
    $mx_subdomain_done = array();
    foreach($this->db->query_fetch_all('select entry_name, entry_value from ' . $this->sql_prefix . 'entrys where entry_zone_id=' . (int)$zone_id . " and entry_type='MX' and entry_name!='' and entry_value like '%.thelounge.net.' order by entry_name;") as $mx_record)
    {
     $data .= $mx_record['entry_name'] . "\t\t";
     if(strlen($mx_record['entry_name']) < 16)
     {
      $data .= "\t";
     }
     if(strlen($mx_record['entry_name']) <= $this->tab_fuzz)
     {
      $data .= "\t";
     }
     $data .= $spf_value . MY_LE;
     $mx_subdomain_done[] = $mx_record['entry_name'];
    }
    /** Einzeleintraege fuer A-Records generieren, nur wenn Domain uebereinstimmt und wir einen A-Entry durchlaufen */
    foreach($spf_liste as $akt_spf)
    {
     if(stristr($akt_spf, '.' . $domain))
     {
      $zsp = str_replace('.' . $domain, '', $akt_spf);
      if($this->is_a_record($domain, $zsp) && !in_array($zsp, $mx_subdomain_done))
      {
       $data .= $akt_spf . ".\t\tIN " . $record_type . "\t\t";
       $data .= '"v=spf1 a' . $spf_string . MY_LE;
      }
     }
    }
   }
   return $data;
  }

  /**
   * TXT-Eintrag fuer die Domain generieren
   *
   * @param  integer $zone_id     Domain-ID
   * @return string
   * @access private
  */
  private function domain_refresh_txt($zone_id)
  {
   /** Domain-Name ermitteln */
   $data = '';
   $domain = $this->domain_get_name_by_id($zone_id);
   /** Alle Eintraege vom Typ "TXT" sortiert nach der Datensatz-Nummer abfragen */
   $result = $this->db->query_fetch_all("select entry_name,entry_value from " . $this->sql_prefix . "entrys where entry_type='TXT' and entry_zone_id=" . (int)$zone_id . ' order by entry_name;', /**$numeric*/true);
   if(count($result) > 0)
   {
    $data .= MY_LE;
    foreach($result as $row)
    {
     switch($row[0])
     {
      case '':
        $data .= '@' . "\t\t\t\tIN TXT\t\t\"" . trim($row[1]) . '"' . MY_LE;
        break;
      default:
        $data .= $row[0] . "\t\t";
        if(strlen($row[0]) < 16)
        {
         $data .= "\t";
        }
        if(strlen($row[0]) <= $this->tab_fuzz)
        {
         $data .= "\t";
        }
        $data .= "IN TXT\t\t\"" . trim($row[1]) . '"' . MY_LE;
        break;
     }
    }
   }
   return $data;
  }

  /**
   * DNS-Delegation-Eintrag fuer die Domain generieren
   *
   * @param  integer $zone_id     Domain-ID
   * @return string
   * @access private
   */
  private function domain_refresh_ns($zone_id)
  {
   /** Domain-Name ermitteln */
   $data = '';
   $domain = $this->domain_get_name_by_id($zone_id);
   /** Alle Eintraege vom Typ "NS" sortiert nach der Datensatz-Nummer abfragen */
   $result = $this->db->query_fetch_all("select entry_name,entry_value from " . $this->sql_prefix . "entrys where entry_type='NS' and entry_zone_id=" . (int)$zone_id . ' order by entry_name;', /**$numeric*/true);
   if(count($result) > 0)
   {
    $data .= MY_LE;
    foreach($result as $row)
    {
     $data .= $row[0] . "\t\t";
     if(strlen($row[0]) < 16)
     {
      $data .= "\t";
     }
     if(strlen($row[0]) <= $this->tab_fuzz)
     {
      $data .= "\t";
     }
     $data .= "IN NS\t\t" . trim($row[1]) . MY_LE;
    }
   }
   return $data;
  }

  /**
   * CNAME-Eintraege fuer die Domain generieren
   *
   * @param  integer $zone_id  Domain-ID
   * @return string
   * @access private
  */
  private function domain_refresh_cname($zone_id)
  {
   $data = '';
   /** Alle Eintrage vom Typ "CNAME" sortiert nach dem Namen abfragen, Eintraege ohne Namen welche es eigentlich nicht geben darf ignorieren */
   $result = $this->db->query_fetch_all("select entry_name, entry_value from " . $this->sql_prefix . "entrys where entry_name!='' and entry_type='CNAME' and entry_zone_id=" . (int)$zone_id . ' order by entry_name;');
   if(count($result) > 0)
   {
    $data .= MY_LE;
    foreach($result as $row)
    {
     $data .= $row['entry_name'] . "\t";
     if(strlen($row['entry_name']) < 27)
     {
      $data .= "\t";
     }
     if(strlen($row['entry_name']) < 16)
     {
      $data .= "\t";
     }
     if(strlen($row['entry_name']) <= $this->tab_fuzz)
     {
      $data .= "\t";
     }
     $data .= 'IN CNAME' . "\t" . $row['entry_value'];
     $data .= MY_LE;
    }
    unset($result);
   }
   return $data;
  }

  /**
   * CNAME/MX-Eintraege die mehr als einen Punkt enthalten und nicht mit
   * einem solchen abschliessen automatisch ergaenzen
   *
   * @param  string $value
   * @return string
   * @access private
  */
  private function check_cname_mx($value)
  {
   if(substr($value, (strlen($value)-1)) != '.' && substr_count($value, '.') > 1)
   {
    $value .= '.';
   }
   return $value;
  }

  /**
   * A-Eintraege die mehr als einen Punkt enthalten und nicht mit
   * einem solchen abschliessen automatisch ergaenzen
   *
   * @param  string $name
   * @return string
   * @access private
  */
  private function check_a_name($name)
  {
   if(substr($name, (strlen($name)-1)) != '.' && substr_count($name, '.') > 1)
   {
    $name .= '.';
   }
   return $name;
  }

  /**
   * Domain-Namen auf formale Gueltigkeit pruefen
   *
   * @param  string $domain
   * @return boolean
   * @todo   Mehr Checks ausser dem nicht erlaubten Punkt am Ende
   * @access private
  */
  private function check_domain_name($domain)
  {
   /** Punkt am Ende ist nicht erlaubt */
   if(substr($domain, (strlen($domain)-1)) == '.')
   {
    return false;
   }
   /** Mindestens 4 Zeichen */
   if(strlen($domain) < 4)
   {
    return false;
   }
   /** Domains aus der Blacklist abweisen */
   if(!empty($this->blacklist))
   {
    if(in_array(strtolower(trim($domain)), $this->blacklist))
    {
     return false;
    }
   }
   /** Wenn alles durchgelaufen -> OK */
   return true;
  }

  /**
   * Zone-File anhand des Domain-Namens oder der Zone-ID auslesen
   *
   * @param  mixed $zone ID oder Name
   * @return string
   * @access public
  */
  public function get_zone_content($zone)
  {
   $row = $this->db->fetch_row($this->db->query('select ' . SQL_BIG_NO_CACHE . ' zone_content from ' . $this->sql_prefix . 'zones where zone_id=' . (int)$zone . " or zone_name='" . $this->db->addslashes($zone) . "';", 1, 0));
   return $row[0];
  }

  /**
   * Eintraege der Tabelle "import" verarbeiten
   * Angegebene Domains loeschen und anhand der vorhandenen Daten neu generieren
   *
   * @access public
   * @return boolean
  */
  public function import()
  {
   /** Abbruch des Skripts verhindern und Batchmode einschalten */
   ignore_user_abort(true);
   set_time_limit(3600);
   $this->batch_mode = true;
   /** Liste der Domains ermitteln */
   $domains = $this->import_domain_list();
   if(!empty($domains))
   {
    foreach($domains as $domain)
    {
     if($this->check_domain_name($domain))
     {
      echo 'Verarbeite Domain &quot;' . rh_htmlentities($domain) . '&quot;<br />' . MY_LE;
      echo 'L&ouml;sche Domain falls bereits vorhanden<br />' . MY_LE;
      $this->domain_remove($domain);
      if($this->domain_add($domain))
      {
       echo '<span style="color:green;">Domain &quot;' . rh_htmlentities($domain) . ' wurde neu angelegt.</span><br />' . MY_LE;
       /** Falls im Import ein A-Entry ohne Name bzw. ein A/CNAME "www" vorhanden ist die automatisch generierten entfernen */
       $this->import_auto_remove($domain);
       /** Eintraege durchlaufen */
       foreach($this->db->query_fetch_all('select ' . SQL_BIG_NO_CACHE . " * from " . $this->sql_prefix . "import where domain='" . $this->db->addslashes($domain) . "' order by name;") as $row)
       {
        extract($row);
        if($this->entry_add($domain, $type, $name, $value, $comment))
        {
         echo '<span style="color:green;">OK - ' . $type . ': ' . $name . ' =&gt; ' . $value . '</span><br />' . MY_LE;
         flush();
        }
        else
        {
         echo '<span style="color:red;">FAILED - ' . $type . ': ' . $name . ' =&gt; ' . $value . '</span><br />' . MY_LE;
         flush();
        }
       }
      }
      else
      {
       echo '<span style="color:red;">Domain &quot;' . rh_htmlentities($domain) . ' konnte nicht angelegt werden, Verarbeitung gestoppt..</span><br />' . MY_LE;
       flush();
      }
      echo '<hr />' . MY_LE;
     }
     else
     {
      echo '<p class="action_failed">Domain &quot;' . rh_htmlentities($domain) . '&quot; ist nicht erlaubt.</p>' . MY_LE;
      flush();
     }
    }
   }
   /** Keine Domains im CSV */
   else
   {
    $this->batch_mode = false;
    return false;
   }
   /** Wenn alles durchgelaufen ist Zones neu generieren */
   $this->batch_mode = false;
   $this->domain_refresh_zones();
   return true;
  }

  /**
   * Liste der Domains aus der Tabelle "Import"
   *
   * @return array
   * @access private
  */
  private function import_domain_list()
  {
   return $this->db->query_fetch_all('select distinct(domain) from ' . $this->sql_prefix . 'import order by domain;', /**$numeric*/true, /**$field*/0);
  }

  /**
   * Automatisch angelegten Default-A und www-CNAME ggf. entfernen
   *
   * @param string $domain
   * @return void
   * @access private
  */
  private function import_auto_remove($domain)
  {
   /** Default A */
   $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "import where type='A' and name='' and domain='" . $this->db->addslashes($domain) . "';", 1, 0));
   if($row[0] > 0)
   {
    $this->db->query("delete from " . $this->sql_prefix . "entrys where entry_type='A' and entry_name='' and entry_zone_cleartext='" . $this->db->addslashes($domain) . "';", 1, 1);
   }
   /** www */
   $row = $this->db->fetch_row($this->db->query("select count(*) from " . $this->sql_prefix . "import where (type='A' or type='CNAME') and name='www' and domain='" . $this->db->addslashes($domain) . "';", 1, 0));
   if($row[0] > 0)
   {
    $this->db->query("delete from " . $this->sql_prefix . "entrys where (entry_type='A' or entry_type='CNAME') and entry_name='www' and entry_zone_cleartext='" . $this->db->addslashes($domain) . "';", 1, 1);
   }
  }

  /**
   * Transaktions-Sperre setzen
   *
   * @return void
   * @access public
  */
  public function gui_lock()
  {
   $this->db->query('update ' . $this->sql_prefix . 'meta set gui_lock=1;', 1, 1);
  }

  /**
   * Sperre aufheben
   *
   * @return void
   * @access public
  */
  public function gui_unlock()
  {
   $this->db->query('update ' . $this->sql_prefix . 'meta set gui_lock=0;', 1, 1);
  }

  /**
   * Abfrage ob Sperre aktiv ist
   *
   * @return boolean
   * @access public
  */
  public function gui_check_lock()
  {
   $row = $this->db->fetch_row($this->db->query('select gui_lock from ' . $this->sql_prefix . 'meta limit 1;', 1, 0));
   if($row[0] == 1)
   {
    echo '<p class="action_failed">Transaktions-Sperre aktiv, bitte sp&auml;ter nochmal versuchen.</p>' . MY_LE;
    return true;
   }
   else
   {
    return false;
   }
  }

  /**
   * Anzahl bestimmter Eintraege einer Domain ermitteln
   *
   * @param  integer $zone_id Domain-ID
   * @param  string  $type A, CNAME, MX, SPF
   * @return integer
   * @access public
  */
  public function entry_count($zone_id, $type)
  {
   $row = $this->db->fetch_row($this->db->query('select ' . SQL_SMALL_CACHE . ' count(*) from ' . $this->sql_prefix . 'entrys where entry_zone_id=' . (int)$zone_id . " and entry_type='" . $this->db->addslashes($type) . "';", 1, 0));
   return $row[0];
  }

  /**
   * Anhand eines Namens ohne Domain den entsprechenden IP-Eintrag der Lounge
   * ermitteln um es zu ermoeglichen bei allen anderen Zonen in der Datenbank
   * die IP-Adressen nicht anzugeben
   *
   * @param  string $name
   * @return string
   * @access public
  */
  public function get_lounge_ip($name)
  {
   $row = $this->db->fetch_row($this->db->query('select entry_value from ' . $this->sql_prefix . "entrys where entry_zone_cleartext='thelounge.net' and entry_type='A' and entry_name='" . $this->db->addslashes($name) . "';", 1, 0));
   if(!empty($row[0]))
   {
    return $row[0];
   }
   else
   {
    return $name;
   }
  }

  /**
   * Serial der angegebenen Zone abrufen
   *
   * @param  string $domain
   * @return string
   * @access public
  */
  public function domain_get_serial($domain)
  {
   $row = $this->db->fetch_row($this->db->query('select zone_serial from ' . $this->sql_prefix . "zones where zone_name='" . $this->db->addslashes($domain) . "' limit 1;", 1, 0));
   return $row[0];
  }

  /**
   * Boolsche Abfrage ob striktes SPF aktiviert ist
   *
   * @param  string $domain
   * @return boolean
   * @access public
  */
  public function domain_get_spf_strict($domain)
  {
   $row = $this->db->fetch_row($this->db->query('select zone_spf_strict from ' . $this->sql_prefix . "zones where zone_name='" . $this->db->addslashes($domain) . "' limit 1;", 1, 0));
   if($row[0])
   {
    return true;
   }
   else
   {
    return false;
   }
  }

  /**
   * SPF-Policy einer Domain setzen
   *
   * @param  string  $domain
   * @param  boolean $value
   * @return boolean
   * @access public
  */
  public function domain_set_spf_strict($domain, $value)
  {
   $rw = $this->db->query('update ' . $this->sql_prefix . 'zones set zone_spf_strict=' . (int)$value . " where zone_name='" . $this->db->addslashes($domain) . "';");
   if($rw)
   {
    $this->domain_refresh_zones($domain);
    $GLOBALS['cl_api']->log('OK: change spf-policy of domain "' . $domain . '" to ' . (int)$value, 'dns');
   }
   return $rw;
  }

  /**
   * Abfrage der aktuellen Client-IO einer Zone
   *
   * @param  string $domain
   * @return integer
   * @access public
  */
  public function domain_get_client_id($domain)
  {
   $row = $this->db->fetch_row($this->db->query('select zone_client_id from ' . $this->sql_prefix . "zones where zone_name='" . $this->db->addslashes($domain) . "' limit 1;", 1, 0));
   return (int)$row[0];
  }

  /**
   * Client-ID einer Domain setzen
   *
   * @param  string  $domain
   * @param  integer $value
   * @return boolean
   * @access public
  */
  public function domain_set_client_id($domain, $value)
  {
   $rw = $this->db->query('update ' . $this->sql_prefix . 'zones set zone_client_id=' . (int)$value . " where zone_name='" . $this->db->addslashes($domain) . "';");
   if($rw)
   {
    $this->domain_refresh_zones($domain);
    $GLOBALS['cl_api']->log('OK: change client-id of domain "' . $domain . '" to ' . (int)$value, 'dns');
   }
   return $rw;
  }

  /**
   * TLL aus Domain-Konfiguration oder globalen Default zurueckgeben
   * @param  string $domain
   * @return integer
   * @access public
  */
  public function domain_get_ttl($domain)
  {
   $row = $this->db->fetch_row($this->db->query('select zone_ttl from ' . $this->sql_prefix . "zones where zone_name='" . $this->db->addslashes($domain) . "' limit 1;", 1, 0));
   settype($row[0], 'integer');
   if($row[0] < 60)
   {
    return $this->ttl;
   }
   else
   {
    return $row[0];
   }
  }

  /**
   * TTL einer Domain setzen
   *
   * @param  string  $domain
   * @param  integer $value
   * @return boolean
   * @access public
  */
  public function domain_set_ttl($domain, $value)
  {
   $rw = $this->db->query('update ' . $this->sql_prefix . 'zones set zone_ttl=' . (int)$value . " where zone_name='" . $this->db->addslashes($domain) . "';");
   if($rw)
   {
    $this->domain_refresh_zones($domain);
    $GLOBALS['cl_api']->log('OK: change TTL of domain "' . $domain . '" to ' . (int)$value, 'dns');
   }
   return $rw;
  }

  /**
   * Whois-Kommando einer Domain mittels "popen" ausfuehren und Antwort zurueckgeben
   *
   * @param  string $domain
   * @access public
   * @return string
  */
  public function whois($domain)
  {
   $out = '';
   /** Kommando mittels "popen" ausfuehren und Rueckgabe in Variable schreiben */
   $temp = popen('/usr/bin/whois -H ' . escapeshellarg($domain) , 'r');
   while($data = fread($temp, 1024))
   {
    $out .= $data;
   }
   pclose($temp);
   /** Ausgabe des Shell-Kommandos als LATIN zurueckgeben */
   return  trim(utf8_decode($out));
  }

  /**
   * Alle aktiven A-Entrys und Hostnamen auflisten
   *
   * @return array
   * @access public
  */
  public function get_all_a_and_cname()
  {
   $arr = array();
   foreach($this->db->query_fetch_all('select' . SQL_BIG_NO_CACHE . ' entry_zone_cleartext, entry_type, entry_name, entry_value from ' . $this->sql_prefix . "entrys where entry_type='A' or entry_type='CNAME' order by entry_zone_cleartext,entry_name;") as $row)
   {
    extract($row);
    switch($entry_type)
    {
     case 'A':
       if(empty($entry_name))
       {
        $arr[] = $entry_zone_cleartext;
       }
       else
       {
        $arr[] = $entry_name . '.' . $entry_zone_cleartext;
       }
       break;
     case 'CNAME':
       $arr[] = $entry_name . '.' . $entry_zone_cleartext;
       break;
    }
   }
   return array_unique($arr);
  }

  /**
   * A-Records ohne MAC-Angabe in der uebergebenen Domain mittels ping/arp durchsuchen
   * und automaisch ergaenzen
   *
   * @param  string $domain
   * @return void
   * @access public
  */
  public function find_mac($domain)
  {
   /** Sicherstellen dass Domain uebergeben wurde */
   $domain = trim($domain);
   if(empty($domain))
   {
    return false;
   }
   /** A-Records ohne MAC abfragen */
   $result = $this->db->query_fetch_all('select entry_id, entry_name, entry_value from ' . $this->sql_prefix . "entrys where entry_mac='' and entry_type='A' and entry_name not like 'z-ptr%' and entry_value!='10.0.0.255' and (entry_value like '10.%' or entry_value like '192.168.%') and entry_zone_cleartext='" . $this->db->addslashes($domain) . "' order by entry_value;");
   if(count($result) > 0)
   {
    $data   = array();
    $xcount = 0;
    foreach($result as $row)
    {
     $data[$xcount] = $row;
     $out = '';
     /** PING mittels "popen" ausfuehren u */
     $temp = popen('/bin/ping -c 1 -W 2 -w 2 -q ' . escapeshellarg($row['entry_value']) , 'r');
     while($buffer = fread($temp, 1024))
     {
      $out .= $buffer;
     }
     pclose($temp);
     /** ARP mittels "popen" ausfuehren und parsen */
     $out = '';
     $temp = popen('/sbin/arp ' . escapeshellarg($row['entry_value']) , 'r');
     while($buffer = fread($temp, 1024))
     {
      $out .= $buffer;
     }
     pclose($temp);
     $out = str_replace(MY_LE, ' ', $out);
     $out = str_replace("\t", ' ', $out);
     $out = preg_replace('/  +|' . "\t" . '/', ' ', $out);
     $out = explode(' ', $out);
     $mac = @$out[8];
     $mac = str_replace('-', ':', strtolower(trim($mac)));
     if(substr_count($mac, ':') < 5)
     {
      $mac = '';
     }
     $data[$xcount]['mac'] = $mac;
     $xcount++;
    }
    unset($result);
    foreach($data as $row)
    {
     if(!empty($row['mac']))
     {
      $this->entry_set_mac($row['entry_id'], $row['mac']);
     }
     echo '<hr />';
    }
    unset($data);
   }
   else
   {
    return false;
   }
  }

  /**
   * Domain-Daten als CSV herunterladen
   *
   * @param  string $domain  Domain-Name
   * @param  string $format  lounge: Interne Referenzen, external: A-Records auf IP-Adressen uebersetzen
   * @return void
   * @access public
  */
  public function export($domain, $format='lounge')
  {
   /** Domain-Daten abfragen */
   $result = $this->db->query_fetch_all('select entry_zone_cleartext as domain, entry_type as type, entry_name as name, entry_value as value, entry_comment as comment, entry_mac as hwaddr from '. $this->sql_prefix . "entrys where entry_zone_cleartext='" . $this->db->addslashes($domain) . "';");
   /** Optional Lounge-A-Records auf IP-Adressen uebersetzen */
   if($format == 'external')
   {
    foreach($result as $key=>$row)
    {
     if($row['type'] == 'A')
     {
      $result[$key]['value'] = $this->get_lounge_ip($row['value']);
     }
    }
   }
   /** Array in CSV konvertieren */
   $csv_lib = new rh_csv();
   $csv_data = $csv_lib->from_array(/**$data*/$result, /**$trenner*/',', /**$text_trenner*/'"', /**$ci*/0, /**$replace_column_names*/array());
   unset($result);
   /** CSV zum Download anbieten */
   ob_end_clean();
   @session_write_close();
   rh_misc::GetMimeType('csv', /**$writeheader*/true);
   header('Content-Disposition: attachment; filename="'. $domain . '.csv"');
   header('Content-Transfer-Encoding: binary');
   exit($csv_data);
  }
 }
?>